from .common_webscraper_functions import fetch_soup_from_page, row_has_link, get_tr_of_stats_table, get_tr_of_table_with_id, get_text_of_element_with_attributes, get_url_of_element_with_attributes, does_html_object_exist, static_value, get_text_of_element_with_type, get_value_from_element
import logging

class CreateFromPageParserFactory:

    def __init__(self, create_from_page_parser_dict) -> None:
        self.create_from_page_parser = CreateFromPageParser(create_from_page_parser_dict['base_url'], DataDictParserFactory(create_from_page_parser_dict['parser']).data_dict_parser)


class CreateFromPageParser:

    def __init__(self, base_url, data_dict_parser) -> None:
        self.base_url = base_url
        self.data_dict_parser = data_dict_parser

    def parse(self, url, webscraperObjectCollection):
        soup = fetch_soup_from_page(self.base_url + url)
        data_dict = self.data_dict_parser.parse(soup, {}, webscraperObjectCollection)
        data_dict['url'] = url
        return data_dict

class TableParserObject:

    def __init__(self, all_object_selection_function, narrow_down_function, data_dict_parser):
        self.all_object_selection_function = all_object_selection_function
        self.narrow_down_function = narrow_down_function
        self.data_dict_parser = data_dict_parser

    def parse_page(self, soup, data_dict, webscraperObjectCollection):
        return_array = []
        for eligible_element in self.all_object_selection_function(soup):
            if self.narrow_down_function(eligible_element):
                return_array.append(self.data_dict_parser.parse(eligible_element, data_dict, webscraperObjectCollection))
        return return_array

class ParserObjectFactory:

    def _get_narrow_down_function(self, function_name):
        if function_name == 'row_has_link':
            return row_has_link
        else:
            raise Exception("No match for narrow down function")


    def __init__(self, parser_dict):
        self.parser_dict = parser_dict
        if parser_dict['parser_type'] == 'table':
            if 'table_id' in parser_dict.keys():
                self.parser = TableParserObject(get_tr_of_table_with_id(parser_dict['table_id']), self._get_narrow_down_function(parser_dict['narrow_down_function']), DataDictParserFactory(parser_dict['data_dict_parser']).data_dict_parser)
            else:
                self.parser = TableParserObject(get_tr_of_stats_table(), self._get_narrow_down_function(parser_dict['narrow_down_function']), DataDictParserFactory(parser_dict['data_dict_parser']).data_dict_parser)
        else:
            raise Exception("No match for parser type")


class DataDictParserFactory:

    def __init__(self, data_dict_parser_dict):
        logging.info('[DataDictParserFactory] [Init] ' + str(data_dict_parser_dict))
        return_dict = {}
        dict_values = {}
        object_urls = []
        object_urls_create_if_not_exist = []
        for field_dict in data_dict_parser_dict:
            if field_dict['parse_type'] == 'input_dict':
                dict_values[field_dict['field_name']] = field_dict['dict_key']
            elif field_dict['parse_type'] == 'url_of_object':
                object_urls.append(field_dict)
            elif field_dict['parse_type'] == 'url_of_object_create_if_not_exist':
                object_urls_create_if_not_exist.append(field_dict)
            else:
                return_dict[field_dict['field_name']] = FunctionParserFactory(field_dict).function
        self.data_dict_parser = DataDictParser(return_dict, dict_values, object_urls, object_urls_create_if_not_exist)

class DataDictParser:

    def __init__(self, function_dict, dict_values, object_urls, object_urls_create_if_not_exist):
        self.function_dict = function_dict
        self.dict_values = dict_values
        self.object_urls = object_urls
        self.object_urls_create_if_not_exist = object_urls_create_if_not_exist

    def parse(self, html_object, data_dict, webscraperObject):
        return_dict = {}
        for key in self.function_dict:
            return_dict[key] = self.function_dict[key](html_object)
        for key in self.dict_values:
            return_dict[key] = data_dict[self.dict_values[key]]
        for object_url in self.object_urls:
            return_dict[object_url['field_name']] = webscraperObject.databaseObject.tables[object_url['object_name']].get_primary_key_by_search_dict({'url': get_url_of_element_with_attributes(object_url['attributes'])(html_object)})
        for object_url in self.object_urls_create_if_not_exist:
            try:
                return_dict[object_url['field_name']] = webscraperObject.databaseObject.tables[object_url['object_name']].get_primary_key_by_search_dict({'url': get_url_of_element_with_attributes(object_url['attributes'])(html_object)})
            except Exception:
                return_dict[object_url['field_name']] = webscraperObject.get_webscraper_object_with_name(object_url['object_name']).create_from_page(get_url_of_element_with_attributes(object_url['attributes'])(html_object), webscraperObject)
        return return_dict


class FunctionParserFactory:

    def _get_function(self, function_name, attributes, field_dict):
        if function_name == 'get_text_of_element_with_attributes':
            if 'remove_strings' in field_dict.keys():
                return get_text_of_element_with_attributes(attributes, field_dict['remove_strings'])
            return get_text_of_element_with_attributes(attributes)
        elif function_name == 'get_url_of_element_with_attributes':
            return get_url_of_element_with_attributes(attributes)
        elif function_name == 'does_html_object_exist':
            return does_html_object_exist(attributes, field_dict['html_object'])
        elif function_name == 'get_text_of_element_with_type':
            return get_text_of_element_with_type(field_dict['element_type'])
        else:
            raise Exception("No match for function " + function_name)

    def __init__(self, field_dict):
        if field_dict['parse_type'] == 'dynamic':
            if 'function_name' in field_dict.keys():
                self.function = self._get_function(field_dict['function_name'], field_dict['attributes'], field_dict)
            else:
                self.function = get_value_from_element(field_dict)
        elif field_dict['parse_type'] == 'static':
            self.function = static_value(field_dict['static_value'])
        else:
            raise Exception("No match for function parse_type")


from bradleystevenson2015_database import database
import json
from .common_webscraper_functions import fetch_soup_from_page
from .parser import CreateFromPageParserFactory, ParserObjectFactory

class WebscraperObjectCollection:

    def __init__(self, webscraper_schema_filepath, database_path, database_schema_filepath, custom_objects):
        self.databaseObject = database.Database(database_path, database_schema_filepath)
        self._create_webscraper_objects(webscraper_schema_filepath, custom_objects)


    def _create_webscraper_objects(self, webscraper_schema_filepath, custom_objects):
        self.webscrapers = []
        file = open(webscraper_schema_filepath)
        data = json.load(file)
        file.close()
        for webscraper_object in data['objects']:
            self.webscrapers.append(WebscraperObjectFactory(webscraper_object, custom_objects).webscraper)


    def get_webscraper_object_with_name(self, object_name):
        for webscraper in self.webscrapers:
            if webscraper.object_name == object_name:
                return webscraper

    def run(self, arguments):
        for webscraper in self.webscrapers:
            if webscraper.object_name in arguments:
                webscraper.create_from_web(self)
            else:
                webscraper.create_from_database(self)
        self.databaseObject.insert_into_database()


class WebscraperObject:

    def __init__(self, object_name, tables, create_from_page_parser=None):
        self.object_name = object_name
        self.tables = tables
        self.create_from_page_parser = create_from_page_parser

    def create_from_web(self, webscraperObjectCollection):
        pass

    def create_from_database(self, webscraperObjectCollection):
        for table_name in self.tables:
            webscraperObjectCollection.databaseObject.tables[table_name].generate_from_database()

    def create_from_page(self, url, webscraperObjectCollection):
        if self.create_from_page_parser is None:
            raise Exception("We have no way to create this object")
        data_dict = self.create_from_page_parser.parse(url, webscraperObjectCollection)
        data_dict['url'] = url
        print(data_dict)
        return webscraperObjectCollection.databaseObject.tables[self.tables[0]].append(data_dict)
    

class WebscraperMultiplePageObject(WebscraperObject):

    def __init__(self, object_name, table, base_url, iterator_table_name, parsers, create_from_page_parser):
        self.base_url = base_url
        self.parsers = parsers
        self.table_name = table
        self.iterator_table_name = iterator_table_name
        super().__init__(object_name, [table], create_from_page_parser)

    def create_from_web(self, webscraperObjectCollection):
        for data_dict in webscraperObjectCollection.databaseObject.tables[self.iterator_table_name].data:
            soup = fetch_soup_from_page(self.base_url + data_dict['url'])
            print(self.base_url + data_dict['url'])
            for parser in self.parsers:
                data = parser.parse_page(soup, data_dict, webscraperObjectCollection)
                for data_dict in data:
                    webscraperObjectCollection.databaseObject.tables[self.table_name].append(data_dict)

class WebscraperStaticPageObject(WebscraperObject):

    def __init__(self, object_name, table, urls, parsers, create_from_page_parser):
        self.urls = urls
        self.parsers = parsers
        self.table_name = table
        super().__init__(object_name, [table], create_from_page_parser)

    def create_from_web(self, webscraperObjectCollection):
        for url in self.urls:
            soup = fetch_soup_from_page(url)
            for parser in self.parsers:
                data = parser.parse_page(soup, {}, webscraperObjectCollection)
                for data_dict in data:
                    webscraperObjectCollection.databaseObject.tables[self.table_name].append(data_dict)

class WebscraperObjectFactory:

    def __init__(self, webscraper_object_dict, custom_objects):
        self.webscrapper_object_dict = webscraper_object_dict
        self.create_from_page_parser = None
        if 'create_from_page_parser' in webscraper_object_dict.keys():
            self.create_from_page_parser =  CreateFromPageParserFactory(webscraper_object_dict['create_from_page_parser']).create_from_page_parser
        if webscraper_object_dict['object_type'] == 'single_page':
            parsers = []
            for parser_dict in self.webscrapper_object_dict['parsers']:
                parsers.append(ParserObjectFactory(parser_dict).parser)
            self.webscraper = WebscraperStaticPageObject(webscraper_object_dict['object_name'], webscraper_object_dict['tables'][0], webscraper_object_dict['urls'], parsers, self.create_from_page_parser)
        elif webscraper_object_dict['object_type'] == 'multiple_page':
            parsers = []
            for parser_dict in self.webscrapper_object_dict['parsers']:
                parsers.append(ParserObjectFactory(parser_dict).parser)
            self.webscraper = WebscraperMultiplePageObject(webscraper_object_dict['object_name'], webscraper_object_dict['tables'][0], webscraper_object_dict['base_url'], webscraper_object_dict['iterator_table_name'], parsers, self.create_from_page_parser)
        elif webscraper_object_dict['object_type'] == 'custom_object':
            for custom_object in custom_objects:
                if custom_object.object_name == webscraper_object_dict['object_name']:
                    self.webscraper = custom_object
        else:
            raise Exception("No match for object type")


from bs4 import BeautifulSoup
from selenium import webdriver
import selenium

def get_element(field_dict):
    element_type = None
    attributes = {}
    if 'element_type' in field_dict.keys():
        element_type = field_dict['element_type']
    if 'attributes' in field_dict.keys():
        attributes = field_dict['attributes']
    if element_type is None:
        def return_function(html_object):
            return html_object.find(attrs=attributes)
        return return_function
    def return_function(html_object):
        return html_object.find(element_type, attrs=attributes)
    return return_function

def get_value_from_element(field_dict):
    if field_dict['html_field_value'] == 'url':
        def return_function(html_object):
            return get_element(field_dict)(html_object).find('a')['href']
        return return_function
    elif field_dict['html_field_value'] == 'text':
        def return_function(html_object):
            return get_element(field_dict)(html_object).text
        return return_function
    else:
        raise Exception(f"No match for html field value {field_dict['html_field_value']}")

def static_value(value):
    def return_function(html_object):
        return value
    return return_function

def does_html_object_exist(attributes, object_type):
    def return_function(html_object):
        if get_element_with_attributes(attributes)(html_object).find(object_type) is not None:
            return 1
        return 0
    return return_function

def get_text_of_element_with_attributes(attributes, remove_strings=[]):
    def return_function(html_object):
        return_string = get_element_with_attributes(attributes)(html_object).text
        for remove_string in remove_strings:
            return_string = return_string.replace(remove_string, '')
        return return_string
    return return_function

def get_url_of_element_with_attributes(attributes):
    def return_function(html_object):
        return get_element_with_attributes(attributes)(html_object).find('a')['href']
    return return_function

def get_text_of_element_with_type(type_name):
    def return_function(html_object):
        return get_element_with_type(type_name)(html_object).text
    return return_function

def get_element_with_type(type_name):
    def return_function(html_object):
        return html_object.find(type_name)
    return return_function

def get_element_with_attributes(attributes):
    def return_function(html_object):
        return html_object.find(attrs=attributes)
    return return_function

def get_tr_of_table_with_id(table_id):
    def return_function(soup):
        return soup.find(id=table_id).find("tbody").find_all("tr")
    return return_function

def get_tr_of_stats_table():
    def return_function(soup):
        return soup.find(attrs={'class': 'stats_table'}).find('tbody').find_all('tr')
    return return_function

def row_has_link(html_object):
    return html_object.find("a") is not None

def fetch_soup_from_page(url):
    while True:
        try:
            options = webdriver.ChromeOptions()
            options.add_argument('headless')
            options.add_argument('window-size=1200x600')
            driver = webdriver.Chrome(options=options)
            driver.get(url)
            page = driver.page_source
            driver.quit()
            soup = BeautifulSoup(page, 'html.parser')
            return soup
        except selenium.common.exceptions.TimeoutException:
            print("Failing url: url")
            print("Timed out loading page, trying again")
        except selenium.common.exceptions.WebDriverException:
            print("failing url: " + str(url))
            print("Web Driver Error, trying again")
